#include <proto.h>

#include "gttags.h"

/************************************************************************\
*
* Module for tags and taglists.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

static GtTag *vt=NULL;
static int vtnr=64,vtcp=0;

/************************************************************************\
*
* Allocate a number of tags.
*
\************************************************************************/
struct GtTag *GtAllocTags(int nr)
{
	if(nr>0)
		return((struct GtTag *)malloc(nr*sizeof(struct GtTag)));
	else
		return(NULL);
}

/************************************************************************\
*
* Free allocated tags.
*
\************************************************************************/
void GtFreeTags(struct GtTag *t)
{
	if(t)
		free(t);
}

/************************************************************************\
*
* Step through a taglist.
*
\************************************************************************/
struct GtTag *GtNextTag(struct GtTag **t)
{
	struct GtTag *tmp;
	
	if(t && *t)
	{
		while(*t)
		{
			tmp=*t;
			switch(tmp->tag)
			{
				case GT_DONE:
					*t=NULL;
					return(NULL);
					break;
				case GT_MORE:
					*t=(GtTag *)tmp->d.ptr;
					break;
				case GT_SKIP:
					*t+=tmp->d.val+1;
					break;
				case GT_IGNORE:
					*t+=1;
					break;
				default:
					*t+=1;
					return(tmp);
					break;
			}
		}
	}
	return(NULL);
}

/************************************************************************\
*
* Copy source taglist to destination taglist.
*
\************************************************************************/
struct GtTag *GtRefreshTags(struct GtTag *dest,struct GtTag *src)
{
	struct GtTag *tmp,*t;
	
	tmp=dest;
	t=src;
	while((src=GtNextTag(&t)))
	{
		dest->tag=src->tag;
		dest->d.val=src->d.val;
		dest++;
	}
	return(tmp);
}

/************************************************************************\
*
* Copy a taglist to a new allocated taglist.
*
\************************************************************************/
struct GtTag *GtCloneTags(struct GtTag *t)
{
	int n=1;
	struct GtTag *tmp;
	
	if(t)
	{
		tmp=t;
		while(GtNextTag(&tmp))
			n++;
		if((tmp=GtAllocTags(n)))
			return(GtRefreshTags(tmp,t));
	}
	return(NULL);
}

/************************************************************************\
*
* Find a given tag in a taglist.
*
\************************************************************************/
struct GtTag *GtFindTag(struct GtTag *t,ulong v)
{
	if(t)
	{
		while(t->tag)
		{
			switch(t->tag)
			{
				case GT_MORE:
					t=(GtTag *)t->d.ptr;
					break;
				case GT_SKIP:
					t+=t->d.val+1;
					break;
				case GT_IGNORE:
					t++;
					break;
				default:
					if(t->tag==v)
						return(t);
					else
						t++;
					break;
			}
		}
	}
	return(NULL);
}

/************************************************************************\
*
* Get long value of a given tag.
*
\************************************************************************/
ulong GtGetTagVal(struct GtTag *t,ulong v,ulong def)
{
	struct GtTag *found;
	
	if((found=GtFindTag(t,v)))
		return(found->d.val);
	else
		return(def);
}

/************************************************************************\
*
* Get pointer value of a given tag.
*
\************************************************************************/
void *GtGetTagPtr(struct GtTag *t,ulong v,void *def)
{
	struct GtTag *found;
	
	if((found=GtFindTag(t,v)))
		return(found->d.ptr);
	else
		return(def);
}

/************************************************************************\
*
* Pack boolean tags into flags.
*
* b - A taglist with flagmask for each boolean tag.
*
\************************************************************************/
ulong GtPackFlagTags(struct GtTag *t,struct GtTag *b,ulong flags)
{
	struct GtTag *c,*f;
	
	while((c=GtNextTag(&t)))
	{
		f=GtFindTag(b,c->tag);
		if(f)
		{
			if(c->d.val)
				flags |= f->d.val;
			else
				flags &= ~f->d.val;
		}
	}
	return(flags);
}


/************************************************************************\
*
* Put tagvalues in an array of longs.
*
* If barray!=NULL set corrsponding boolean to true if value found.
* off - Taglist with index as tagvalues.
*
\************************************************************************/
void GtArrayTags(struct GtTag *t,struct GtTag *off,ulong *array,ubyte *barray)
{
	struct GtTag *c,*f;
	
	while((c=GtNextTag(&t)))
	{
		f=GtFindTag(off,c->tag);
		if(f)
		{
			array[f->d.val]=c->d.val;
			if(barray)
				barray[f->d.val]=1;
		}
	}
}

/************************************************************************\
*
* Change tags in a list according to a maplist. Return number of changes.
*
\************************************************************************/
int GtMapTags(struct GtTag *t,struct GtTag *map)
{
	struct GtTag *c,*f;
	int n=0;
	
	while((c=GtNextTag(&t)))
	{
		f=GtFindTag(map,c->tag);
		if(f)
		{
			c->tag=f->d.val;
			n++;
		}
	}
	return(n);
}

/************************************************************************\
*
* Begin vararg tags.
*
\************************************************************************/
int GtBeginVaTags(void)
{
	return(vtcp);
}

/************************************************************************\
*
* End vararg tags.
*
\************************************************************************/
void GtEndVaTags(int p)
{
	vtcp=p;
}

/************************************************************************\
*
* Copy vararg tags to array.
*
\************************************************************************/
GtTag *GtCopyVaTags(ulong first,va_list *va)
{
	int n=0;
	
	if(!vt)
	{
		if(!(vt=(GtTag *)malloc(sizeof(GtTag)*vtnr)))
			return(NULL);
		memset(vt,0,sizeof(GtTag)*vtnr);
	}
	while(1)
	{
		if(n+vtcp>vtnr)
		{
			vtnr+=64;
			if(!(vt=(GtTag *)realloc(vt,sizeof(GtTag)*vtnr)))
				return(NULL);
		}
		if(n==0)
			vt[vtcp].tag=first;
		else
			vt[vtcp].tag=va_arg(*va,ulong);
		if((vt[vtcp].tag!=GT_DONE) && (vt[vtcp].tag!=GT_MORE))
			vt[vtcp].d.val=va_arg(*va,ulong);
		else
		{
			vtcp++;
			return(&vt[vtcp-n-1]);
		}
		n++;
		vtcp++;
	}
	return(NULL);
}
